#include	<P16F628.INC>

; RA0/AN0 - sensor 1
; RA1/AN1 - sensor 2
; RA2/AN2 - reference
; RA3     - charge
; RA4     - unused
; RA5     - button
; RA6     - 4094 CLOCK
; RA7     - 4094 DATA
; RB0     - A1-A3/E4-E6
; RB1     - F1-F3/D4-D6
; RB2     - H1-H6
; RB3     - B1-B3/C4-C6
; RB4     - C1-C3/G4-G6
; RB5     - G1-G3/B4-B6
; RB6     - D1-D3/F4-F6
; RB7     - E1-E3/A4-A6

; Digits order: 4 1 2 3 6 5

	list		p=16f628,f=INHX32
	__config	_INTRC_OSC_NOCLKOUT & _PWRTE_OFF & _MCLRE_OFF & _BODEN_ON & _LVP_OFF


;MCOUNT		equ		.30000		; measure cycles count (max 32767)
MSTIME		equ		.100


#define bank0 bcf	STATUS,RP0
#define bank1 bsf	STATUS,RP0

W_TEMP		equ	0x70
W_TEMP2		equ	0x71
CSPTR		equ	0x72
FSR_TEMP2	equ	0x73
TIMR3		equ	0x74


STATUS_TEMP	equ	0x20
C1			equ 0x21
VRAM		equ 0x22 ; 0x22-0x27
VDIG6		equ	0x22
VDIG3		equ	0x23
VDIG2		equ	0x24
VDIG1		equ	0x25
VDIG4		equ	0x26
VDIG5		equ	0x27
VPTR		equ 0x28
FSR_TEMP	equ	0x29
PA			equ	0x2A
KTIMER		equ	0x2B
KSTATE		equ	0x2C
TTIMER		equ	0x2D
TTIM2		equ	0x2E
FLASH		equ	0x2F
FMASK		equ	0x30
T0			equ	0x31
T1			equ	0x32

D1			equ	0x33
D2			equ	0x34
D3			equ	0x35
D4			equ	0x36
DCOUNT		equ	0x37

LAST1P		equ	0x3A	; (2)
LAST1M		equ	0x3C	; (2)
LAST2P		equ	0x3E	; (2)
LAST2M		equ	0x40	; (2)

AADR		equ	0x42
CADR		equ 0x43

CSEL		equ	0x44
LASTVL		equ	0x45
LASTVH		equ	0x46
CFLAGS		equ	0x47
CCN2		equ	0x48

CCP1		equ	0x49
CCP2		equ	0x4A
CCP			equ	0x4B

TM1			equ 0x50
TM2			equ 0x51

R1e			equ	0x52
R1s			equ	0x53
R1d			equ	0x54
R1c			equ	0x55
R1b			equ	0x56
R1a			equ	0x57

R2e			equ	0x58
R2s			equ	0x59
R2d			equ	0x5A
R2c			equ	0x5B
R2b			equ	0x5C
R2a			equ	0x5D

R3d			equ	0x5E
R3c			equ	0x5F
R3b			equ	0x60
R3a			equ	0x61

R4c			equ	0x62
R4b			equ	0x63
R4a			equ	0x64

RT1			equ	0x65
RT2			equ	0x66
RT3			equ	0x67


CSTACK		equ	0xA0

bsf1	macro	reg,bit
	bsf	(reg^0x80),bit
    endm
bcf1	macro	reg,bit
	bcf	(reg^0x80),bit
    endm
movwf1	macro	arg
    movwf	(arg^0x80)
    endm
movf1	macro	arg,d
    movf	(arg^0x80),d
    endm
clrf1	macro	arg
    clrf	(arg^0x80)
    endm

check_int	macro	endl
	local	ci1,ci2
	btfsc	PIR1,TMR2IF			; 1
	goto	ci1					; 2 / 3
	goto	$+1					; 4 / -
	goto	ci2					; 6 / -
ci1	bank1						; - / 4
	bsf1	TRISA,3				; - / 5
	bsf		INTCON,GIE			; 
	nop							;
	btfsc	TIMR3,7
	goto	endl
	bcf		INTCON,GIE			;
	bcf1	TRISA,3				;
	bank0						; 6 / 6
ci2
	endm
;	---------------------
cmp16i	macro	VAR,value
		movf	VAR+1,W
		sublw	HIGH value
		btfss	STATUS,Z
		goto	$+3
		movf	VAR,W
		sublw	LOW value
		endm
;	---------------------
	org	00000h	; Reset Vector
  	goto	start
;	---------------------------
;	Interrupt
	org 	4		
	movwf	W_TEMP
	swapf	STATUS,W
	bank0
	movwf	STATUS_TEMP
	movf	FSR,W
	movwf	FSR_TEMP

	incf	VPTR,F
	movf	VPTR,W
	sublw	0x05
	btfsc	STATUS,C
	goto	NO_RS
	clrf	VPTR
	bsf		PA,7
NO_RS
	movf	VPTR,W
	addlw	VRAM
	movwf	FSR
	clrf	PORTB

	movf	PA,W
	iorlw	0x40
	movwf	PORTA
	andlw	0xBF
	movwf	PORTA
	
	movf	INDF,W
	andwf	FMASK,W
	movwf	PORTB
	bcf		PA,7

	; Timeout timer
	incfsz	TTIM2,F
	goto	STTIM
	movlw	.1
	xorwf	FLASH,F
	movlw	0xFF
	btfsc	STATUS,Z
	movlw	0xFB
	movwf	FMASK
	movf	TTIMER,F
	btfsc	STATUS,Z
	goto	STTIM
	decfsz	TTIMER,F
	goto	STTIM
	bsf		KSTATE,4		; timeout
STTIM


	movf	VPTR,F
	btfss	STATUS,Z
	goto	SKIP_KEY

	incf	TIMR3,F

	; check button state
	incf	KTIMER,W
	btfss	STATUS,Z
	movwf	KTIMER
	movf	KTIMER,W
	sublw	.12
	btfsc	STATUS,C
	goto	SKIP_KEY
	movf	PORTA,W
	xorwf	KSTATE,W
	andlw	0x20
	btfsc	STATUS,Z
	goto	NO_CHG
	; button state changed
	xorwf	KSTATE,F
	btfss	KSTATE,5
	goto	KEY_PRESSED
	; button released
	btfsc	KSTATE,2
	bsf		KSTATE,6		; long press released
	bcf		KSTATE,2		; clear long press
	movf	KTIMER,W
	sublw	.50
	btfss	STATUS,C
	goto	KEY_PRESSED
	; short press detected
	btfsc	KSTATE,3
	goto	KEY_DBL
	bsf		KSTATE,3
	goto	KEY_PRESSED
KEY_DBL
	; double press detected
	bsf		KSTATE,1
	bcf		KSTATE,3
KEY_PRESSED
	clrf	KTIMER
	goto	SKIP_KEY

NO_CHG
	; button state not changed
	movf	KTIMER,W
	btfss	KSTATE,5
	goto	KEY_HOLD
	; button is released
	btfss	KSTATE,3
	goto	SKIP_KEY
	sublw	.80
	btfsc	STATUS,C
	goto	SKIP_KEY
	; single press detected
	bsf		KSTATE,0
	bcf		KSTATE,3
	goto	SKIP_KEY
KEY_HOLD
	; button is hold
	btfsc	KSTATE,2
	goto	SKIP_KEY
	sublw	.80
	btfss	STATUS,C
	bsf		KSTATE,2	; long press detected
SKIP_KEY
TIM1
	movf	FSR_TEMP,W
	movwf	FSR
	bcf		PIR1,TMR2IF
	swapf	STATUS_TEMP,W
	movwf	STATUS
	swapf	W_TEMP,F
	swapf	W_TEMP,W
	retfie
;	------------------------
start
	banksel	0
	clrf	PA
	clrf	TMR2
	clrf	PORTA
	clrf	PORTB
	; Init comparator
	movlw	(1<<CM1)|(1<<CM0)
	movwf	CMCON

	bank1
	clrf1	PIE1
	movlw	b'00111111'
	movwf1	TRISA
	clrf1	TRISB

;	clear 4094
	bank0
	movlw	6
	movwf	C1
C1LOOP
	bsf		PORTA,6
	nop
	bcf		PORTA,6
	decfsz	C1,f
	goto	C1LOOP

	clrf	VPTR
	movlw	.2
	movwf	FLASH
	movlw	0x20
	movwf	KSTATE
	movlw	0xFF
	movwf	FMASK
	movwf	KTIMER
	clrf	TTIMER
	clrf	TTIM2
	call	CSCLEAR
;	--------------
	movlw	HIGH DIG2IND1
	movwf	PCLATH

	; Init Timer
	movlw	(1<<TMR2ON)|(1<<T2CKPS0)
	movwf	T2CON
	bank1
	movlw	.249
	movwf1	PR2
	movlw	b'11011111'
	movwf1	OPTION_REG
	movlw	(1<<TMR2IE)
	movwf1	PIE1
	movlw	(1<<GIE)|(1<<PEIE)
	movwf	INTCON

	clrwdt
	bank0
	movlw	0xFF
	movwf	VDIG1
	movwf	VDIG2
	movwf	VDIG3
	movwf	VDIG4
	movwf	VDIG5
	movwf	VDIG6

	movlw	0x20
	andwf	KSTATE,F
	movlw	.8
	movwf	TTIMER
it1	clrwdt
	btfsc	KSTATE,4
	goto	STARTMAIN	; timeout
	btfss	KSTATE,1
	goto	it1			
	goto	CALIBR		; double click
;	--------------
STARTMAIN
	clrf	VDIG1
	clrf	VDIG2
	clrf	VDIG3
	clrf	VDIG4
	clrf	VDIG5
	clrf	VDIG6

	bsf		FLASH,1

	; Measure loop
	;bank0
ML0
	call	MPRINT1
	call	MPRINT2

	btfsc	KSTATE,5
	clrwdt
	goto	ML0
;	-----------------
;	Calibrate dialog
CALIBR
	; Read current points number
	movlw	LOW CP1
	call	EEREAD
	movwf	CCP1
	sublw	.9
	btfss	STATUS,C
	clrf	CCP1

	movlw	LOW CP2
	call	EEREAD
	movwf	CCP2
	sublw	.9
	btfss	STATUS,C
	clrf	CCP2

	bcf		FLASH,1

CL0a
	movlw	.3
	movwf	CSEL

	; Select channels
CL0
	movlw	0x20
	andwf	KSTATE,F

	clrf	VDIG1
	clrf	VDIG2
	clrf	VDIG3
	btfss	CSEL,1
	goto	CL1
	movlw	.12
	call	DIG2IND1
	movwf	VDIG3
	movf	CCP1,W
	call	DIG2IND1
	movwf	VDIG1
CL1
	clrf	VDIG4
	clrf	VDIG5
	clrf	VDIG6
	btfss	CSEL,0
	goto	CL2
	movlw	.12
	call	DIG2IND2
	movwf	VDIG6
	movf	CCP2,W
	call	DIG2IND2
	movwf	VDIG4

	movlw	.40
	movwf	TTIMER
	bcf		KSTATE,4
CL2
	clrwdt
	btfsc	KSTATE,4
	goto	STARTMAIN		; timeout
	btfsc	KSTATE,1
	goto	CAL_ENTER		; double click
	btfsc	KSTATE,6
	goto	RES_PTS
	btfss	KSTATE,0
	goto	CL2
	; single click - change channels selection
	bcf		KSTATE,0
	decfsz	CSEL,F
	goto	CL0
	goto	CL0a

RES_PTS
	; Reset points
	btfsc	CSEL,1
	clrf	CCP1
	btfsc	CSEL,0
	clrf	CCP2
	goto	CL0

CAL_ENTER
	movlw	0x20
	andwf	KSTATE,F
CL3
	clrwdt
	movf	TTIMER,F
	btfsc	CSEL,1
	call	MPR1a
	btfsc	KSTATE,1	; double click - return unchanged
	goto	CL0
	btfsc	KSTATE,2
	goto	CAL_SLIDE
	btfsc	CSEL,0
	call	MPR2a
	btfsc	KSTATE,1	; double click - return unchanged
	goto	CL0
	btfsc	KSTATE,2
	goto	CAL_SLIDE
	goto	CL3

CAL_SLIDE
	clrf	CFLAGS
	clrf	TIMR3
	clrf	CCN2
SL0
	movlw	.40
	movwf	TTIMER
	btfss	CSEL,1
	goto	SL1
	call	lv2r1
	call	PR16_1
SL1
	btfss	CSEL,0
	goto	SL2
	call	lv2r1
	call	PR16_2
SL2
	clrwdt
	btfsc	KSTATE,0
	goto	SL5
	btfss	KSTATE,6
	goto	SL4
	clrf	CCN2			; long press released
SL5
	movlw	0x01			; change slide direction
	xorwf	CFLAGS,F
	bcf		KSTATE,6
	bcf		KSTATE,0
SL4
	btfsc	KSTATE,4
	goto	CAL_ENTER		; timeout - exit unchanged
	btfsc	KSTATE,1
	goto	CAL_SET			; double click - set value
	btfss	KSTATE,2
	goto	SL2
	movlw	.4
	btfsc	CCN2,5
	goto	SL6
	addlw	.8
	btfsc	CCN2,4
	goto	SL6
	addlw	.16
	btfss	CCN2,3
	addlw	.32
SL6
	subwf	TIMR3,W
	btfss	STATUS,C
	goto	SL2
	clrf	TIMR3
	btfss	CCN2,5
	incf	CCN2,F			; accelerate
	
	btfsc	LASTVH,7
	goto	SL7
	cmp16i	LASTVL,.1000
	movlw	.1
	btfss	STATUS,Z
	goto	SL6a
	btfss	CFLAGS,0
	goto	SET10
	goto	SL8
SL6a
	btfss	STATUS,C
	goto	SET10
	goto	SL8

SL7
	cmp16i	LASTVL,(.65536-.100)
	movlw	.1
	btfss	STATUS,Z
	goto	SL7a
	btfsc	CFLAGS,0
	goto	SET10
	goto	SL8
SL7a	
	btfsc	STATUS,C
SET10
	movlw	.10
SL8
	movwf	T0
	btfsc	CFLAGS,0
	goto	SL3

	btfsc	LASTVH,7
	goto	SL8a
	cmp16i	LASTVL,.1499
	btfss	STATUS,C
	goto	SL0
SL8a
	movf	T0,W
	addwf	LASTVL,F
	btfss	STATUS,C
	goto	SL0
	incf	LASTVH,F
	goto	SL0

SL3
	btfss	LASTVH,7
	goto	SL8b
	cmp16i	LASTVL,(.65536-.800)
	btfsc	STATUS,C
	goto	SL0
SL8b
	movf	T0,W
	sublw	0
	addwf	LASTVL,F
	btfsc	STATUS,C
	goto	SL0
	decf	LASTVH,F
	goto	SL0

CAL_SET
	movlw	0x20
	andwf	KSTATE,F
	
	btfss	CSEL,1
	goto	SKIP_S1
	incf	CCP1,W
	movwf	CCP
	movwf	CCP1

	call	CSCLEAR
	call	LOADMR1
	movlw	LOW CP1
	call	CALC3

SKIP_S1
	btfss	CSEL,0
	goto	SKIP_S2
	incf	CCP2,W
	movwf	CCP
	movwf	CCP2

	call	CSCLEAR
	call	LOADMR2
	movlw	LOW CP2
	call	CALC3

SKIP_S2
	goto	CL0
;	-----------------
lv2r1
	movf	LASTVL,W
	movwf	R1a
	movf	LASTVH,W
	movwf	R1b
	clrf	R1c
	clrf	R1d
	clrf	R1s
	btfss	LASTVH,7
	return
	bsf		R1s,7
	comf	R1a,F
	comf	R1b,F
	incfsz	R1a,F
	return
	incf	R1b,F
	return
;	-----------------
MPR1a
	call	CSCLEAR
	movf	CCP1,F
	btfss	STATUS,Z
	goto	mpr1b
	call	LOADB
	call	LOADA
	goto	mpr1

mpr1b
	decfsz	CCP1,W
	goto	mpr1a
	call	LOADB
	movlw	LOW CC1 + 4
	call	ELOAD32
	goto	mpr1

MPRINT1
	call	CSCLEAR
mpr1a
	movlw	LOW CC1
	call	ELOAD64
mpr1
	call	MEASURE1
	btfsc	STATUS,C
	goto	ERR1
	call	LOADMR1
	call	CALCT
	btfsc	STATUS,C
	goto	ERR1
PR16_1
	; Print R1 on ind 1
	call	convert_to_digits
	movf	INDF,W
	call	DIG2IND1
	movwf	VDIG3
	decf	FSR,F
	movf	INDF,W
	call	DIG2IND1
	btfsc	INDF,7
	iorlw	0x04
	movwf	VDIG2
	decf	FSR,F
	movf	INDF,W
	call	DIG2IND1
	btfsc	INDF,7
	iorlw	0x04
	movwf	VDIG1
	return
;	-----------------
MPR2a
	call	CSCLEAR
	movf	CCP2,F
	btfss	STATUS,Z
	goto	mpr2b
	call	LOADB
	call	LOADA
	goto	mpr2

mpr2b	
	decfsz	CCP2,W
	goto	mpr2a
	call	LOADB
	movlw	LOW CC2 + 4
	call	ELOAD32
	goto	mpr2

MPRINT2
	call	CSCLEAR
mpr2a
	movlw	LOW CC2
	call	ELOAD64
mpr2
	call	MEASURE2
	btfsc	STATUS,C
	goto	ERR2
	call	LOADMR2
	call	CALCT
	btfsc	STATUS,C
	goto	ERR2

PR16_2
	; Print R1 on ind 2
	call	convert_to_digits
	movf	INDF,W
	call	DIG2IND2
	movwf	VDIG6
	decf	FSR,F
	movf	INDF,W
	call	DIG2IND2
	btfsc	INDF,7
	iorlw	0x04
	movwf	VDIG5
	decf	FSR,F
	movf	INDF,W
	call	DIG2IND2
	btfsc	INDF,7
	iorlw	0x04
	movwf	VDIG4
	return
;	-----------------
;	Calculate temperature
CALCT
	call	FDIV			; p/m
	call	FLN				; ln(p/m)
	call	FSUB			; A-ln(..)
	call	FDIV			; B/(A-ln(..))
	call	LOADT0			; 273.15
	call	FSUB			; t

	movlw	.10
	call	FPUSHB
	call	FMUL
	call	FPOP24

	movf	R1a,W
	movwf	LASTVL
	movf	R1b,W
	movwf	LASTVH
	btfss	R1s,7
	return
	comf	LASTVH,F
	comf	LASTVL,F
	incfsz	LASTVL,F
	return
	incf	LASTVH,F
	return
;	----------------------
LOADT0
	; Load constant: 273.15
	movf	CSPTR,F
	btfsc	STATUS,Z
	return
	movlw	0x66
	call	PUSHW
	movlw	0x26
	call	PUSHW
	movlw	0x11
	call	PUSHW
	movlw	0x48
	goto	PUSHW
;	----------------------
LOADMR1
	movf	LAST1M,W
	movwf	R1b
	movf	LAST1M+1,W
	movwf	R1c
	call	FPUSH16
	movf	LAST1P,W
	movwf	R1b
	movf	LAST1P+1,W
	movwf	R1c
	goto	FPUSH16
;	----------------------
LOADMR2
	movf	LAST2M,W
	movwf	R1b
	movf	LAST2M+1,W
	movwf	R1c
	call	FPUSH16
	movf	LAST2P,W
	movwf	R1b
	movf	LAST2P+1,W
	movwf	R1c
	goto	FPUSH16
;	----------------------
LOADA
	; Load default value for A (11.6237)
	movf	CSPTR,F
	btfsc	STATUS,Z
	return
	movlw	0x5A
	call	PUSHW
	movlw	0xF5
	call	PUSHW
	movlw	0x73
	call	PUSHW
	movlw	0x43
	goto	PUSHW
;	----------------------
LOADB
	; Load default value for B (3175)
	movf	CSPTR,F
	btfsc	STATUS,Z
	return
	movlw	0x00
	call	PUSHW
	movlw	0xE0
	call	PUSHW
	movlw	0x8C
	call	PUSHW
	movlw	0x4B
	goto	PUSHW
;	----------------------
CALC3
	movwf	CADR
	movwf	AADR
	; m p
	call	FDIV
	call	FLN				; d = ln(p/m)

	; Accumulate data
	call	FCOPY
	; d d

	movlw	.1
	call	FPUSHB
	movf	LASTVL,W
	movwf	R1b
	movf	LASTVH,W
	movwf	R1c
	call	FPUSH16S		; t*10
	movlw	.10
	call	FPUSHB
	call	FDIV			; t
	call	LOADT0			; 273.15
	call	FADD			; T
	call	FDIV			; f = 1/T
	; f d d
	
	movf	CSPTR,F
	btfsc	STATUS,Z
	return

	movf	CCP,W
	call	EWRITE

	call	FCOPY
	; f f d d
	decf	CCP,W
	btfsc	STATUS,Z
	goto	cc1
	call	ELOAD32C
	call	FADD
cc1	call	ESAVE32		; f += f[i]
	
	; f d d
	call	FMUL
	; fd d
	decf	CCP,W
	btfsc	STATUS,Z
	goto	cc2
	call	ELOAD32C
	call	FADD
cc2	call	ESAVE32		; fd += fd[i]

	; d	
	call	FCOPY		; d[i]
	; d d
	decf	CCP,W
	btfsc	STATUS,Z
	goto	cc3
	call	ELOAD32C
	call	FADD
cc3	call	ESAVE32		; d += d[i]
	
	; d
	call	FCOPY
	call	FMUL		; d[i]^2
	; d^2
	decf	CCP,W
	btfsc	STATUS,Z
	goto	cc4
	call	ELOAD32C
	call	FADD
cc4	call	ESAVE32		; d2 += d[i]^2
	
	decf	CCP,W
	btfss	STATUS,Z
	goto	cc5
	; Single point

	;movf	CADR,W
	;call	ELOAD32		; B
	call	LOADB
	movlw	.1
	addwf	AADR,W
	call	ELOAD32		; f
	call	FMUL
	movlw	.9
	addwf	AADR,W
	call	ELOAD32		; d
	call	FADD
	call	LOADB		; default value for B
	call	ESAVE32
	call	ESAVE32
	return

cc5
	; Multiple points
	movlw	.13
	addwf	AADR,W
	call	ELOAD32		; d2	(K21)
	movf	CCP,W
	call	FPUSHB
	call	FMUL
	; d2*n

	movlw	.9
	addwf	AADR,W
	call	ELOAD32		; d		(G27)
	call	FCOPY
	call	FMUL		; d^2
	call	FSUB		; d2*n - d^2 (H29)
	; H29

	call	FCOPY
	; H29 H29

	movlw	.1
	addwf	AADR,W
	call	ELOAD32		; f (J21)
	movlw	.9
	addwf	AADR,W
	call	ELOAD32		; d (I21)
	call	FMUL
	; G33 H29 H29
	movlw	.5
	addwf	AADR,W
	call	ELOAD32		; fd (H21)
	; H21 G33 H29 H29
	movf	CCP,W
	call	FPUSHB
	call	FMUL		; fd*n
	; G34 G33 H29 H29
	call	FSUB		; G33-G34 (H34)
	; H34 H29 H29
	call	FDIV		; H29/H34 (G36)
	; G36 H29
	call	FCOPY
	call	ESAVE32

	movlw	.1
	addwf	AADR,W
	call	ELOAD32		; f	(J21)
	; J21 G36 H29
	movlw	.13
	addwf	AADR,W
	call	ELOAD32		; d2 (K21)
	; K21 J21 G36 H29
	call	FMUL
	; G25 G36 H29
	movlw	.5
	addwf	AADR,W
	call	ELOAD32		; fd	(H21)
	; H21 G25 G36 H29
	movlw	.9
	addwf	AADR,W
	call	ELOAD32		; d		(I21)
	; I21 H21 G25 G36 H29
	call	FMUL
	; G24 G25 G36 H29
	call	FSUB		; G25-G24	(H25)
	; H25 G36 H29
	call	FMUL
	; H25*G36 H29
	call	FXCHG
	call	FDIV
	; H37 (=G37)
	call	ESAVE32
	return	
;	----------------------
EEREAD
	bank1
	errorlevel	-302
	movwf	EEADR
	bsf		EECON1,RD
	movf	EEDATA,W
	errorlevel	+302
	bank0
	return
;	----------------------
MEASURE1
	bcf		INTCON,GIE

	clrf	TIMR3
	bank1
	bcf1	TRISA,3
	bank0

	btfsc	CMCON,C1OUT
	goto	C1IS1
	bsf		PA,3
	movf	PA,W
	movwf	PORTA
	bank1
	bcf1	TRISA,3
	bank0
C1WT0
	check_int	MSERR
	btfss	CMCON,C1OUT
	goto	C1WT0
C1IS1
	bcf		PA,3
	movf	PA,W
	movwf	PORTA
	bank1
	bcf1	TRISA,3
	bank0
C1WT1
	check_int	MSERR
	btfsc	CMCON,C1OUT
	goto	C1WT1

	movlw	.127-MSTIME
	movwf	TIMR3
	clrf	LAST1P
	clrf	LAST1P+1
	clrf	LAST1M
	clrf	LAST1M+1
	bsf		PA,3
	movf	PA,W
	movwf	PORTA

C1ML
	check_int	C1END		; 6
	btfsc	CMCON,C1OUT		; 7 / 8
	goto	C1SET0			; 8
	goto	$+1
	goto	$+4
C1SET1
	bsf		PA,3
	movf	PA,W
	movwf	PORTA
	incf	LAST1P,F		;   /  8
	btfsc	STATUS,Z		;   /  9 / 10
	incf	LAST1P+1,F		;   / 10
	goto	C1ML

C1ML2
	check_int	C1END
	btfss	CMCON,C1OUT
	goto	C1SET1
	goto	$+1
	goto	$+4
C1SET0
	bcf		PA,3
	movf	PA,W
	movwf	PORTA
	incf	LAST1M,F
	btfsc	STATUS,Z
	incf	LAST1M+1,F
	goto	C1ML2

C1END
	bank1
	bsf1	TRISA,3
	bank0
	bcf		STATUS,C
	bsf		INTCON,GIE
	return

MSERR
	bank0
	bsf		STATUS,C
	return
;	----------------------
MEASURE2
	bcf		INTCON,GIE

	clrf	TIMR3
	bank1
	bcf1	TRISA,3
	bank0

	btfsc	CMCON,C2OUT
	goto	C2IS1
	bsf		PA,3
	movf	PA,W
	movwf	PORTA
	bank1
	bcf1	TRISA,3
	bank0
C2WT0
	check_int	MSERR
	btfss	CMCON,C2OUT
	goto	C2WT0
C2IS1
	bcf		PA,3
	movf	PA,W
	movwf	PORTA
	bank1
	bcf1	TRISA,3
	bank0
C2WT1
	check_int	MSERR
	btfsc	CMCON,C2OUT
	goto	C2WT1

	movlw	.127-MSTIME
	movwf	TIMR3
	clrf	LAST2P
	clrf	LAST2P+1
	clrf	LAST2M
	clrf	LAST2M+1
	bsf		PA,3
	movf	PA,W
	movwf	PORTA

C2ML
	check_int	C2END		; 6
	btfsc	CMCON,C2OUT		; 7 / 8
	goto	C2SET0			; 8
	goto	$+1
	goto	$+4
C2SET1
	bsf		PA,3
	movf	PA,W
	movwf	PORTA
	incf	LAST2P,F		;   /  8
	btfsc	STATUS,Z		;   /  9 / 10
	incf	LAST2P+1,F		;   / 10
	goto	C2ML

C2ML2
	check_int	C2END
	btfss	CMCON,C2OUT
	goto	C2SET1
	goto	$+1
	goto	$+4
C2SET0
	bcf		PA,3
	movf	PA,W
	movwf	PORTA
	incf	LAST2M,F
	btfsc	STATUS,Z
	incf	LAST2M+1,F
	goto	C2ML2

C2END
	bank1
	bsf1	TRISA,3
	bank0
	bcf		STATUS,C
	bsf		INTCON,GIE
	return
;	----------------------
;	Clear stack
CSCLEAR
	movlw	CSTACK-1
	movwf	CSPTR
	swapf	CSPTR,F
	return
;	----------------------
CSENTER			; set FSR to stack top
	movwf	W_TEMP2
	swapf	W_TEMP2,F
	swapf	FSR,W
	movwf	FSR_TEMP2
	swapf	CSPTR,W
	movwf	FSR
	swapf	W_TEMP2,W
	return

CSLEAVE			; save and restore FSR
	movwf	W_TEMP2
	swapf	W_TEMP2,F
	swapf	FSR,W
	movwf	CSPTR
	swapf	FSR_TEMP2,W
	movwf	FSR
	swapf	W_TEMP2,W
	return
;	----------------------
csenter	macro
		movf	CSPTR,F
		btfsc	STATUS,Z
		return
		call	CSENTER
		endm
;	----------------------
FPUSH16S
	movlw	0x50
	clrf	R1s
	btfss	R1c,7
	goto	fp8
	bsf		R1s,7
	comf	R1c,F
	comf	R1b,F
	incfsz	R1b,F
	goto	fp8
	incf	R1c,F
	goto	fp8

FPUSH16
	movlw	0x50
	goto	fp1
;	Push unsigned byte as integer
FPUSHB
	movwf	R1c
	clrf	R1b
	movlw	0x48
fp1	clrf	R1s
fp8	clrf	R1a
	clrf	R1d
	movwf	R1e
	csenter
	goto	nret1
;	----------------------
FCOPY
	csenter
	call	inr1
	call	outr1
	goto	retr1
;	----------------------
FXCHG
	csenter
	call	inr1
	call	inr2
	call	outr1
	call	outr2
	goto	CSLEAVE
;	----------------------
;	POP value to R1 and round it to integer
FPOP24
	bsf		STATUS,C
	csenter
	call	inr1
	movf	R1e,W
	btfsc	STATUS,Z
	goto	fp6
	sublw	0x40 + .24
	btfss	STATUS,C
	goto	fp5
	movwf	RT1
	btfsc	STATUS,Z
	goto	fp3
	sublw	.25
	btfss	STATUS,C
	goto	fp6
fp4	bcf		STATUS,C
	rrf		R1d,F
	rrf		R1c,F
	rrf		R1b,F
	rrf		R1a,F
	decfsz	RT1,F
	goto	fp4
fp3	btfss	STATUS,C
	goto	fp2
	incfsz	R1a,F
	goto	fp2
	incfsz	R1b,F
	goto	fp2
	incfsz	R1c,F
	goto	fp2
fp5	; Number too fit
	bsf		STATUS,C
	goto	CSLEAVE

fp6	; Zero
	clrf	R1a
	clrf	R1b
	clrf	R1c
	clrf	R1s
fp2
fp7	bcf		STATUS,C
	goto	CSLEAVE
;	----------------------
FSUB
	csenter
	movlw	0x80
	xorwf	INDF,F
	goto	fa0

FADD
	csenter
fa0	movf	INDF,W
	andlw	0x7F
	movwf	RT1
	swapf	FSR,W
	movwf	RT2
	movlw	.4
	subwf	FSR,F
	movf	INDF,W
	andlw	0x7F
	subwf	RT1,W
	swapf	RT2,W
	movwf	FSR
	btfsc	STATUS,C
	goto	fa1
	call	inr1
	call	inr2
	goto	fa2

fa1	call	inr2
	call	inr1

fa2	
	; Scale exponents
	movf	R1e,W
	btfsc	STATUS,Z
	goto	fab
	subwf	R2e,F
	btfsc	STATUS,Z
	goto	fa4
fa3	bcf		STATUS,C
	rrf		R1d,F
	rrf		R1c,F
	rrf		R1b,F
	rrf		R1a,F
	incf	R1e,F
	decfsz	R2e,F
	goto	fa3
fa4

	; Compare signs
	movf	R1s,W
	xorwf	R2s,W
	btfss	STATUS,Z
	call	negr1
fa9	; Do ADD
	call	addr12
	; Normalize R1
nret1
	btfsc	R1d,7
	call	negr1
	movf	R1d,F
	bcf		STATUS,C
	btfss	STATUS,Z
	goto	fa6
	movlw	.24
	movwf	RT1
fa8	decf	R1e,F
	btfsc	STATUS,Z
	goto	fa7
	bcf		STATUS,C
	rlf		R1a,F
	rlf		R1b,F
	rlf		R1c,F
	btfsc	STATUS,C
	goto	fa7
	decfsz	RT1,F
	goto	fa8
	clrf	R1e
	clrf	R1s
	goto	fa7

fa5	bcf		STATUS,C
	rrf		R1d,F
	rrf		R1c,F
	rrf		R1b,F
	rrf		R1a,F
	incf	R1e,F
fa6	decfsz	R1d,W
	goto	fa5
	btfsc	STATUS,C
	call	incr1
fa7
retr1
	call	outr1
	goto	CSLEAVE
fab	call	outr2
	goto	CSLEAVE
;	----------------------
;	R1 += R2
addr12
	movf	R2a,W
	addwf	R1a,F
	btfsc	STATUS,C
	call	incr1b
	movf	R2b,W
	addwf	R1b,F
	btfsc	STATUS,C
	call	incr1c
	movf	R2c,W
	addwf	R1c,F
	btfsc	STATUS,C
	incf	R1d,F
	movf	R2d,W
	addwf	R1d,F
	return
;	----------------------
inr1
	clrf	R1d
	clrf	R1s
	rlf		INDF,W
	rrf		R1s,F
	movf	INDF,W
	andlw	0x7F
	btfss	STATUS,Z
	incf	R1d,F
	movwf	R1e
	decf	FSR,F
	movf	INDF,W
	movwf	R1c
	decf	FSR,F
	movf	INDF,W
	movwf	R1b
	decf	FSR,F
	movf	INDF,W
	movwf	R1a
	decf	FSR,F
	return
;	----------------------
inr2
	clrf	R2d
	clrf	R2s
	rlf		INDF,W
	rrf		R2s,F
	movf	INDF,W
	andlw	0x7F
	btfss	STATUS,Z
	incf	R2d,F
	movwf	R2e
	decf	FSR,F
	movf	INDF,W
	movwf	R2c
	decf	FSR,F
	movf	INDF,W
	movwf	R2b
	decf	FSR,F
	movf	INDF,W
	movwf	R2a
	decf	FSR,F
	return
;	----------------------
outr1
	incf	FSR,F
	movf	R1a,W
	movwf	INDF
	incf	FSR,F
	movf	R1b,W
	movwf	INDF
	incf	FSR,F
	movf	R1c,W
	movwf	INDF
	incf	FSR,F
	movf	R1s,W
	andlw	0x80
	iorwf	R1e,W
	movwf	INDF
	return
;	----------------------
outr2
	incf	FSR,F
	movf	R2a,W
	movwf	INDF
	incf	FSR,F
	movf	R2b,W
	movwf	INDF
	incf	FSR,F
	movf	R2c,W
	movwf	INDF
	incf	FSR,F
	movf	R2s,W
	andlw	0x80
	iorwf	R2e,W
	movwf	INDF
	return
;	----------------------
negr1
	comf	R1a,F
	comf	R1b,F
	comf	R1c,F
	comf	R1d,F
	movlw	0x80
	xorwf	R1s,F
incr1
	incfsz	R1a,F
	return
incr1b
	incfsz	R1b,F
	return
incr1c
	incfsz	R1c,F
	return
incr1d
	incf	R1d,F
	return
;	----------------------
decr1b
	decf	R1b,F
	incfsz	R1b,W
	return
decr1c
	decf	R1c,F
	incfsz	R1c,W
	return
decr1d
	decf	R1d,F
	return
;	----------------------
incr4b
	incfsz	R4b,F
	return
	incf	R4c,F
	return
;	----------------------
;	Change sign
FNEG
	call	CSENTER
	movlw	0x80
	xorwf	INDF,F
	goto	CSLEAVE
;	----------------------
;	Multiply
FMUL
	csenter
	call	inr2
	call	inr1
	movf	R1e,F
	btfsc	STATUS,Z
	goto	ret0
	movf	R2e,F
	btfsc	STATUS,Z
	goto	ret0
	movf	R1a,W
	movwf	R3a
	movf	R1b,W
	movwf	R3b
	movf	R1c,W
	movwf	R3c
	call	mulr23
	movf	R2e,W
	addlw	0xC0
	addwf	R1e,F
	movf	R2s,W
	xorwf	R1s,F
	goto	nret1

ret0
	clrf	R1a
	clrf	R1b
	clrf	R1c
	clrf	R1d
	clrf	R1e
	clrf	R1s
	goto	retr1
;	----------------------
;	R1 = R2 * R3
mulr23
	movf	R2a,W
	movwf	R1a
	movf	R2b,W
	movwf	R1b
	movf	R2c,W
	movwf	R1c
	movf	R2d,W
	movwf	R1d
	movlw	.24
	movwf	RT1
fm1	bcf		STATUS,C
	rrf		R2d,F
	rrf		R2c,F
	rrf		R2b,F
	rrf		R2a,F
	rlf		R3a,F
	rlf		R3b,F
	rlf		R3c,F
	btfsc	STATUS,C
	call	addr12
	decfsz	RT1,F
	goto	fm1
	return
;	----------------------
;	Divide
FDIV
	csenter
	call	inr2
	call	inr1
	movf	R2e,F
	btfsc	STATUS,Z
	goto	ferr
	movf	R1e,F
	btfsc	STATUS,Z
	goto	ret0
	movlw	.27
	movwf	RT1
	clrf	R3a
	goto	fd4

fd5	bcf		STATUS,C
	rlf		R1a,F
	rlf		R1b,F
	rlf		R1c,F
	rlf		R1d,F
fd4	call	cmpsub
	rlf		R3a,F
	rlf		R3b,F
	rlf		R3c,F
	rlf		R3d,F
	decfsz	RT1,F
	goto	fd5

	movf	R3d,W
	movwf	R1d
	movf	R3c,W
	movwf	R1c
	movf	R3b,W
	movwf	R1b
	movf	R3a,W
	movwf	R1a
	movf	R2e,W
	addlw	0xC2
	subwf	R1e,F
	movf	R2s,W
	xorwf	R1s,F
	goto	nret1

ferr
	clrf	FSR
	goto	CSLEAVE
;	----------------------
;	if (R1 >= R2) {
;		R1 -= R2; 
;		return 1;
;	} else {
;		return 0;
;	}
cmpsub
	movf	R2d,W
	subwf	R1d,W
	btfss	STATUS,Z
	goto	fd2
cmps1
	movf	R2c,W
	subwf	R1c,W
	btfss	STATUS,Z
	goto	fd2
	movf	R2b,W
	subwf	R1b,W
	btfss	STATUS,Z
	goto	fd2
	movf	R2a,W
	subwf	R1a,W
fd2	btfss	STATUS,C
	return
	movf	R2a,W
	subwf	R1a,F
	btfss	STATUS,C
	call	decr1b
	movf	R2b,W
	subwf	R1b,F
	btfss	STATUS,C
	call	decr1c
	movf	R2c,W
	subwf	R1c,F
	btfss	STATUS,C
	call	decr1d
	movf	R2d,W
	subwf	R1d,F
	bsf		STATUS,C
	return
;	----------------------
FLN
	csenter
	call	inr1
	movf	R1e,F
	btfsc	STATUS,Z
	goto	ferr
	movf	R1s,F
	btfss	STATUS,Z
	goto	ferr
	clrf	R2d
	clrf	R4a
	clrf	R4b
	clrf	R4c

fl3	clrf	RT1
	movf	R1a,W
	movwf	R2a
	movf	R1b,W
	movwf	R2b
	movf	R1c,W
	movwf	R2c
	bsf		STATUS,C
fl2	rrf		R2c,F
	rrf		R2b,F
	rrf		R2a,F
	call	cmps1
	btfsc	STATUS,C
	goto	fl1
	incf	RT1,F
	movlw	.24
	subwf	RT1,W
	btfss	STATUS,C
	goto	fl2

	movlw	0x30
	movwf	R3a
	movlw	0xE4
	movwf	R3b
	movlw	0x62
	movwf	R3c
	clrf	R2a
	clrf	R2b
	bcf		STATUS,C
	rlf		R1e,F
	rlf		R1e,W
	movwf	R2c
	clrf	R2d
	rlf		R2d,F
	call	mulr23
	bcf		STATUS,C
	movlw	0xD0
	movwf	R2a
	movlw	0x1B
	movwf	R2b
	movlw	0x9D
	movwf	R2c
	movlw	0xFE
	movwf	R2d
	call	addr12
	movlw	.5
	movwf	RT1
fl4	bcf		STATUS,C
	rlf		R1a,F
	rlf		R1b,F
	rlf		R1c,F
	rlf		R1d,F
	decfsz	RT1,F
	goto	fl4
	movlw	0x40
	movwf	R1e
	movf	R4a,W
	movwf	R2a
	movf	R4b,W
	movwf	R2b
	movf	R4c,W
	movwf	R2c
	clrf	R2d
	goto	fa9

fl1	movf	RT1,W
	call	LNT1
	addwf	R4a,F
	btfsc	STATUS,C
	call	incr4b
	movf	RT1,W
	call	LNT2
	addwf	R4b,F
	btfsc	STATUS,C
	incf	R4c,F
	movf	RT1,W
	call	LNT3
	addwf	R4c,F
	goto	fl3
;	----------------------

;	Push zero
PUSH0
	clrw
;	Push value from W into stack
;	FSR is unchanged
PUSHW
	csenter
	incf	FSR,F
	movwf	INDF
	goto	CSLEAVE
;	----------------------
;   R1 / R2a -> R1
;	R1 % R2a -> R1c
div8
	clrf	R3a
	clrf	R3b
	incf	R3a,F
	;bcf		STATUS,C
dv1	rlf		R1a,F
	rlf		R1b,F
	rlf		R1c,F
	rlf		R1d,F
	btfsc	R1d,0
	goto	dv2
	movf	R2a,W
	subwf	R1c,W
	btfss	STATUS,C
	goto	dv3
dv2	movf	R2a,W
	subwf	R1c,F
	btfss	STATUS,C
	decf	R1d,F
	bsf		STATUS,C
	btfsc	R1d,0
	return					; divide error (C=1)
dv3	rlf		R3a,F
	rlf		R3b,F
	btfss	STATUS,C
	goto	dv1
	movf	R3a,W
	movwf	R1a
	movf	R3b,W
	movwf	R1b
	bcf		STATUS,C		; C=0 (no error)
	return
;	----------------------
;   Load 32-bit integer from EEPROM, addr = W
;	and push it into stack
ELOAD64
	movwf	T0
	movlw	.8
	goto	el3
ELOAD32C
	movf	CADR,W
ELOAD32
	movwf	T0
	movlw	.4
el3	movwf	T1
	csenter
el2	movf	T0,W
	call	EEREAD
	incf	FSR,F
	movwf	INDF
	incf	T0,F
	decfsz	T1,F
	goto	el2
	goto	CSLEAVE
;	----------------------
EWRITE
	bank1
	errorlevel	-302
	movwf	EEDATA
	errorlevel	+302
	bank0
	movf	CADR,W
	incf	CADR,F
	bank1
	errorlevel	-302
	movwf	EEADR
	bsf		EECON1,WREN
	bcf		INTCON,GIE
	movlw	0x55
	movwf	EECON2
	movlw	0xAA
	movwf	EECON2
	bsf		EECON1,WR
	bsf		INTCON,GIE
es1	btfsc	EECON1,WR
	goto	es1
	bcf		EECON1,WREN
	errorlevel	+302
	bank0
	return
;	----------------------
;	POP 32-bit from stack to EEPROM, addr = CADR
;	CADR += 4
ESAVE32
	movlw	.4
es3	csenter
	movwf	T0
	subwf	FSR,W
	movwf	FSR
	movwf	T1
es2	incf	FSR,F
	movf	INDF,W
	call	EWRITE
	decfsz	T0,F
	goto	es2
	movf	T1,W
	movwf	FSR
	goto	CSLEAVE
;	----------------------
convert_to_digits
	movlw	D1
	movwf	FSR
	movlw	.10
	movwf	R2a
	movlw	.4
	movwf	RT3
print_loop
	call	div8
	btfsc	STATUS,C
	goto	print_over
	movf	R1c,W
	movwf	INDF
	clrf	R1c
	incf	FSR,F
	decfsz	RT3,F
	goto	print_loop

	movf	R1a,W
	iorwf	R1b,W
	btfss	STATUS,Z
	goto	print_over

	; Set decimal point placement
	movlw	0x80
	iorwf	D2,F

	; Set minus sign
	decf	FSR,F
	movf	R1s,F
	btfsc	STATUS,Z
	goto	print_plus

	movlw	0x0A
	movf	INDF,F
	btfss	STATUS,Z
	goto	print_over		; overflow
	movf	D3,F
	btfsc	STATUS,Z
	decf	FSR,F
	movwf	INDF
	return

print_plus
	movf	INDF,F
	btfss	STATUS,Z
	return
	decf	FSR,F
	movf	INDF,F
	btfss	STATUS,Z
	return
	movlw	0x0B
	movwf	INDF
	return

print_over
	movlw	0x0A
	movwf	D4
	movlw	0x0B
	movwf	D3
	movwf	D2
	movlw	D4
	movwf	FSR
	return
;	----------------------
;	print error: '---'
ERR1
	movlw	.10
	call	DIG2IND1
	movwf	VDIG3
	movwf	VDIG2
	movwf	VDIG1
	return
ERR2
	movlw	.10
	call	DIG2IND2
	movwf	VDIG6
	movwf	VDIG5
	movwf	VDIG4
	return
; ======================================
	org		0x7A0
LNT3
	addlw	-.8
	btfsc	STATUS,C
	retlw	0x00
	addlw	.8
	addwf	PCL,F
	retlw	0xB1
	retlw	0x49
	retlw	0x22
	retlw	0x10
	retlw	0x08
	retlw	0x04
	retlw	0x02
	retlw	0x01
LNT2
	addlw	-.16
	btfsc	STATUS,C
	retlw	0x00
	addlw	.16
	addwf	PCL,F
	retlw	0x72
	retlw	0xA5
	retlw	0x2F
	retlw	0x85
	retlw	0x20
	retlw	0x08
	retlw	0x02
	retlw	0x00
	retlw	0x80
	retlw	0x40
	retlw	0x20
	retlw	0x10
	retlw	0x08
	retlw	0x04
	retlw	0x02
	retlw	0x01
LNT1
	addlw	-.25
	btfsc	STATUS,C
	retlw	0x00
	addlw	.25
	addwf	PCL,F
	retlw	0x18
	retlw	0x88
	retlw	0x1D
	retlw	0x99
	retlw	0xAF
	retlw	0x16
	retlw	0x03
	retlw	0x80
	retlw	0x20
	retlw	0x08
	retlw	0x02
	retlw	0x01
	retlw	0x00
	retlw	0x00
	retlw	0x00
	retlw	0x00
	retlw	0x80
	retlw	0x40
	retlw	0x20
	retlw	0x10
	retlw	0x08
	retlw	0x04
	retlw	0x02
	retlw	0x01
	retlw	0x01	
;	----------------------
	org		0x7E0
DIG2IND1
	andlw	0x0F
	addwf	PCL,F
	retlw	b'11011011'	;0
	retlw	b'00011000'	;1
	retlw	b'11101001'	;2
	retlw	b'01111001'	;3
	retlw	b'00111010'	;4
	retlw	b'01110011'	;5
	retlw	b'11110011'	;6
	retlw	b'00011001'	;7
	retlw	b'11111011'	;8
	retlw	b'01111011'	;9
	retlw	b'00100000'	;-
	retlw	b'00000000'	;space
	retlw	b'10101011'	;P
;	----------------------
	org		0x7F0
DIG2IND2
	andlw	0x0F
	addwf	PCL,F
	retlw	b'11101011'	;0
	retlw	b'00101000'	;1
	retlw	b'10110011'	;2
	retlw	b'10111010'	;3
	retlw	b'01111000'	;4
	retlw	b'11011010'	;5
	retlw	b'11011011'	;6
	retlw	b'10101000'	;7
	retlw	b'11111011'	;8
	retlw	b'11111010'	;9
	retlw	b'00010000'	;-
	retlw	b'00000000'	;space
	retlw	b'11110001'	;P
;	----------------------
	org		0x2100
	DE		0
CP1		DE	0
CP1A	DE	0,0,0,0
CP1B	DE	0,0,0,0
CP1C	DE	0,0,0,0
CP1D	DE	0,0,0,0
CC1		DE	0x00,0xE0,0x8C,0x4B		; 3175
		DE	0x5A,0xF5,0x73,0x43		; 11.6237
CP2		DE	0
CP2A	DE	0,0,0,0
CP2B	DE	0,0,0,0
CP2C	DE	0,0,0,0
CP2D	DE	0,0,0,0
CC2		DE	0x00,0xE0,0x8C,0x4B		; 3175
		DE	0x5A,0xF5,0x73,0x43		; 11.6237
;	----------------------
	end
